(************************* RANDPORT.MOD ***************************)
MODULE RandPort;
(*
** Filename: RANDPORT.MOD (Modula-2)
**
** Purpose: Random number generator ISAAC (Indirection, Shift,
** Accumulate, Add, and Count), as developed by Bob Jenkins.
**
** This module is optimized for speed except it's portable to
** platforms where integers are bigger than 4 bytes or characters
** are bigger than 1 byte.
**
** This code is a translation of RANDPORT.C
** by Gregory S.Vigneault, z180@hotmail.com
**
** MODIFIED:
**  960327: Creation (addition of randinit, really).
**  970719: use context, not global variables, FOR internal state.
**  980324: make a portable version.
**  981228: translated from C to Modula-2 by G.S.Vigneault.
**
*)
FROM Standard IMPORT RANDSIZ, RANDSIZL, randctx, randctxPTR;

FROM BitOp  IMPORT ANDL, ANDW, SHLL, SHRL, XORL;
FROM InOut  IMPORT Write, WriteHex, WriteLn;
FROM SYSTEM IMPORT ADR;

CONST MASK32 = 4294967295;    (* 0FFFFFFFFH (32-bits) *)

(*----------------------------------------------------------------*)

PROCEDURE isaac (ctx :randctxPTR);
  CONST INDC = LONG(RANDSIZ-1);
  VAR a,b,x,y :LONGCARD;
      m,mm,m2,r,mend :CARDINAL;

      PROCEDURE ind (x :LONGCARD) :LONGCARD;
        BEGIN
          RETURN ctx^.randmem[mm + SHORT( ANDL(SHRL(x,2), INDC) )];
        END ind;

      PROCEDURE rngstep (mix :LONGCARD);
        BEGIN
          x := ctx^.randmem[m];
          a := ANDL( XORL(a,mix) + ctx^.randmem[m2], MASK32 );
          INC(m2);
          y := ANDL( a + b + ind(x), MASK32 );
          ctx^.randmem[m] := y;
          INC(m);
          b := ANDL( x + ind(SHRL(y,RANDSIZL)), MASK32 );
          ctx^.randrsl[r] := b;
          INC(r);
        END rngstep;

  BEGIN (*isaac*)

    mm := 0;    (* index FOR ctx^.randmem[] *)
    r := 0;     (* index FOR ctx^.randrsl[] *)

    a := ctx^.randa;
    INC(ctx^.randc);
    b := ANDL( ctx^.randb + ctx^.randc,  MASK32 );

    m := mm;
    m2 := m + RANDSIZ DIV 2;
    mend := m2;

    WHILE m < mend DO
      rngstep( SHLL(a,13) );
      rngstep( SHRL(a, 6) );
      rngstep( SHLL(a, 2) );
      rngstep( SHRL(a,16) );
    END;

    m2 := mm;

    WHILE m2 < mend DO
      rngstep( SHLL(a,13) );
      rngstep( SHRL(a, 6) );
      rngstep( SHLL(a, 2) );
      rngstep( SHRL(a,16) );
    END;

    ctx^.randb := b;
    ctx^.randa := a;

  END isaac;

(*----------------------------------------------------------------*)

PROCEDURE randinit (ctx :randctxPTR; flag :BOOLEAN);
  (* IF flag=TRUE use contents of randrsl[] to initialize mm[] *)
  VAR a,b,c,d,e,f,g,h :LONGCARD;
      i,m,r :CARDINAL;

      PROCEDURE mix;
        BEGIN
          a := XORL(a, SHLL(b,11));   d := d + a;   b := b + c;
          b := XORL(b, SHRL(c, 2));   e := e + b;   c := c + d;
          c := XORL(c, SHLL(d, 8));   f := f + c;   d := d + e;
          d := XORL(d, SHRL(e,16));   g := g + d;   e := e + f;
          e := XORL(e, SHLL(f,10));   h := h + e;   f := f + g;
          f := XORL(f, SHRL(g, 4));   a := a + f;   g := g + h;
          g := XORL(g, SHLL(h, 8));   b := b + g;   h := h + a;
          h := XORL(h, SHRL(a, 9));   c := c + h;   a := a + b;
        END mix;

  BEGIN (*randinit*)

    ctx^.randa := 0L;
    ctx^.randb := 0L;
    ctx^.randc := 0L;

    m := 0;   (* index FOR ctx^.randmem[] *)
    r := 0;   (* index FOR ctx^.randrsl[] *)

    a := 2654435769;          (* 9E3779B9H: The Golden Ratio *)
    b := a;  c := a;  d := a;
    e := a;  f := a;  g := a;
    h := a;

    FOR i := 0 TO 3 DO  (* scramble it *)
      mix;
    END;

    IF flag THEN

      (* initialize using the contents of r[] as the seed *)

      FOR i := 0 TO RANDSIZ-1 BY 8 DO
        a := a + ctx^.randrsl[i  ];
        b := b + ctx^.randrsl[i+1];
        c := c + ctx^.randrsl[i+2];
        d := d + ctx^.randrsl[i+3];
        e := e + ctx^.randrsl[i+4];
        f := f + ctx^.randrsl[i+5];
        g := g + ctx^.randrsl[i+6];
        h := h + ctx^.randrsl[i+7];
        mix;
        ctx^.randmem[i  ] := a;
        ctx^.randmem[i+1] := b;
        ctx^.randmem[i+2] := c;
        ctx^.randmem[i+3] := d;
        ctx^.randmem[i+4] := e;
        ctx^.randmem[i+5] := f;
        ctx^.randmem[i+6] := g;
        ctx^.randmem[i+7] := h;
      END;

      (* 2nd pass makes all of seed affect all of ctx^.randmem[] *)

      FOR i := 0 TO RANDSIZ-1 BY 8 DO
        a := a + ctx^.randmem[i  ];
        b := b + ctx^.randmem[i+1];
        c := c + ctx^.randmem[i+2];
        d := d + ctx^.randmem[i+3];
        e := e + ctx^.randmem[i+4];
        f := f + ctx^.randmem[i+5];
        g := g + ctx^.randmem[i+6];
        h := h + ctx^.randmem[i+7];
        mix;
        ctx^.randmem[i  ] := a;
        ctx^.randmem[i+1] := b;
        ctx^.randmem[i+2] := c;
        ctx^.randmem[i+3] := d;
        ctx^.randmem[i+4] := e;
        ctx^.randmem[i+5] := f;
        ctx^.randmem[i+6] := g;
        ctx^.randmem[i+7] := h;
      END;

    ELSE

      FOR i := 0 TO RANDSIZ-1 BY 8 DO   (* fill in messy stuff *)
        mix;
        ctx^.randmem[i  ] := a;
        ctx^.randmem[i+1] := b;
        ctx^.randmem[i+2] := c;
        ctx^.randmem[i+3] := d;
        ctx^.randmem[i+4] := e;
        ctx^.randmem[i+5] := f;
        ctx^.randmem[i+6] := g;
        ctx^.randmem[i+7] := h;
      END;

    END; (*IF flag*)

    isaac(ctx);   (* fill in the first set of results *)

    ctx^.randcnt := LONG(RANDSIZ); (* prep for first results *)

  END randinit;

(*================================================================*)

VAR

  i, j :CARDINAL;
  ctx :randctx;

BEGIN (*RandPort*)

  ctx.randa := 0L;
  ctx.randb := 0L;
  ctx.randc := 0L;

  FOR i := 0 TO 255 DO
    ctx.randrsl[i] := 0L;
  END;

  randinit( ADR(ctx), TRUE );

  FOR i := 0 TO 1 DO
    isaac( ADR(ctx) );
    FOR j := 0 TO 255 DO
      WriteHex( SHORT( ctx.randrsl[j] DIV 65536L ), 4 );
      WriteHex( SHORT( ctx.randrsl[j] MOD 65536L ), 4 );
      IF ANDW(j,7) = 7 THEN
        WriteLn;
      END;
    END;
  END;


END RandPort.
(************************* RANDPORT.MOD ***************************)

