(************************* READABLE.MOD ***************************)
MODULE Readable;
(*
** Filename: READABLE.MOD (Modula-2)
**
** Purpose: Random number generator ISAAC (Indirection, Shift,
** Accumulate, Add, and Count), as developed by Bob Jenkins.
**
** This module produces the same results as RAND.MOD but the code
** is easier to read.
**
** This code is a translation of READABLE.C
** by Gregory S. Vigneault, z180@hotmail.com
**
*)
FROM InOut IMPORT Write, WriteHex, WriteLn;
FROM BitOp IMPORT ANDW, SHLL, SHRL, XORL;

VAR

  (* external results... *)

  randcnt : LONGCARD;
  randrsl : ARRAY [0..255] OF LONGCARD;

  (* internal state... *)

  aa,
  bb,
  cc  : LONGCARD;
  mm  : ARRAY [0..255] OF LONGCARD;

(*----------------------------------------------------------------*)

PROCEDURE isaac;
  VAR x, y :LONGCARD;
      i :CARDINAL;
  BEGIN

    INC(cc);          (* cc gets incremented once per 256 results *)
    bb := bb + cc;    (* and then combined with bb *)

    FOR i := 0 TO 255 DO

      x := mm[i];

      CASE (i MOD 4) OF
        0 : aa := XORL(aa, SHLL(aa,13));
      | 1 : aa := XORL(aa, SHRL(aa, 6));
      | 2 : aa := XORL(aa, SHLL(aa, 2));
      | 3 : aa := XORL(aa, SHRL(aa,16));
      END;

      aa := aa + mm[(i + 128) MOD 256];
      y := aa + bb + mm[SHORT( SHRL(x,2) MOD 256L )];
      mm[i] := y;
      bb := x + mm[SHORT( SHRL(y,10) MOD 256L )];
      randrsl[i] := bb;

(*
** Note that bits 2..9 are chosen from x but 10..17 are chosen
** from y.  The only important thing here is that 2..9 and 10..17
** don't overlap.  2..9 and 10..17 were then chosen FOR speed in
** the optimized version (rand.c)
**
** See http://ourworld.compuserve.com/homepages/bob_jenkins/isaac.htm
** for further explanations and analysis.
*)
    END; (*FOR i*)
  END isaac;

(*----------------------------------------------------------------*)

PROCEDURE randinit (flag :BOOLEAN);
  VAR i : CARDINAL;
      a,b,c,d,e,f,g,h : LONGCARD;

      PROCEDURE mix;
        BEGIN
          a := XORL(a, SHLL(b,11));   d := d + a;   b := b + c;
          b := XORL(b, SHRL(c, 2));   e := e + b;   c := c + d;
          c := XORL(c, SHLL(d, 8));   f := f + c;   d := d + e;
          d := XORL(d, SHRL(e,16));   g := g + d;   e := e + f;
          e := XORL(e, SHLL(f,10));   h := h + e;   f := f + g;
          f := XORL(f, SHRL(g, 4));   a := a + f;   g := g + h;
          g := XORL(g, SHLL(h, 8));   b := b + g;   h := h + a;
          h := XORL(h, SHRL(a, 9));   c := c + h;   a := a + b;
        END mix;

  BEGIN (*randinit*)

    aa := 0L;  bb := 0L;  cc := 0L;

    a := 2654435769;  (* 9E3779B9H: The Golden Ratio *)

    b := a;  c := a;  d := a;  e := a;
    f := a;  g := a;  h := a;

    FOR i := 0 TO 3 DO  (* scramble it *)
      mix;
    END;

    FOR i := 0 TO 255 BY 8 DO   (* fill in mm[] with messy stuff *)

      IF flag THEN  (* use all info in the seed *)
        a := a + randrsl[i  ];
        b := b + randrsl[i+1];
        c := c + randrsl[i+2];
        d := d + randrsl[i+3];
        e := e + randrsl[i+4];
        f := f + randrsl[i+5];
        g := g + randrsl[i+6];
        h := h + randrsl[i+7];
      END;

      mix;

      mm[i  ] := a;
      mm[i+1] := b;
      mm[i+2] := c;
      mm[i+3] := d;
      mm[i+4] := e;
      mm[i+5] := f;
      mm[i+6] := g;
      mm[i+7] := h;

    END; (*FOR i*)

    IF flag THEN  (* 2nd pass to make all seed affect all mm[]  *)

      FOR i := 0 TO 255 BY 8 DO
        a := a + mm[i  ];
        b := b + mm[i+1];
        c := c + mm[i+2];
        d := d + mm[i+3];
        e := e + mm[i+4];
        f := f + mm[i+5];
        g := g + mm[i+6];
        h := h + mm[i+7];

        mix;

        mm[i  ] := a;
        mm[i+1] := b;
        mm[i+2] := c;
        mm[i+3] := d;
        mm[i+4] := e;
        mm[i+5] := f;
        mm[i+6] := g;
        mm[i+7] := h;
      END;

    END; (*IF flag*)

    isaac;            (* fill in the first set of results         *)
    randcnt := 256L;  (* prepare to use the first set of results  *)

  END randinit;

(*================================================================*)

VAR i, j :CARDINAL;

BEGIN (*Readable*)

  aa := 0L;  bb := 0L;  cc := 0L;

  FOR i := 0 TO 255 DO
    mm[i] := 0L;
    randrsl[i] := 0L;
  END;

  randinit(TRUE);

  (* produce output equivalent to contents of RANDVECT.TXT ... *)

  FOR i := 0 TO 1 DO
    isaac;
    FOR j := 0 TO 255 DO
      WriteHex( SHORT( randrsl[j] DIV 65536L ), 4 );
      WriteHex( SHORT( randrsl[j] MOD 65536L ), 4 );
      IF ANDW(j,7) = 7 THEN
        WriteLn;
      END;
    END;
  END;

END Readable.
(************************* READABLE.MOD ***************************)

